#
# Author:
#   David Cattermole (cattermole91@gmail.com)
#   
# Description:
#   This script will re-parent an object, keeping the world space
#   position the same except moving the object in to a different
#   hierarchy. This will work with animation data, not just a
#   single frame, like the native Maya command "parent".
# 
# Usage:
#  import reparent
#  reparent.gui()
#
#  Or to do the reparent from the command line:
#  import reparent
#  reparent.main( [ 'theChild1', 'theChild2' ], 'theParent', sparseKeys=True )
#


import maya.cmds as cmds
import maya.mel as mel

reparent_cmd = ( 'import maya.cmds as cmds;\n'
                 'import reparent;\n'
                 'children = cmds.textFieldButtonGrp("reparent_children_textfield", query=True, text=True);\n'
                 'parent = cmds.textFieldButtonGrp("reparent_parent_textfield", query=True, text=True);\n'
                 'children = reparent.splitChildrenString(children);\n'
                 'parent = reparent.splitParentString(parent);\n'
                 'sparseKeys = cmds.checkBoxGrp("reparent_sparseKeys_chkbox", query=True, value1=True);\n'
                 'reparent.main(children, parent, sparseKeys=sparseKeys);\n' )

reset_cmd = ( 'import maya.cmds as cmds;'
              'children = cmds.textFieldButtonGrp("reparent_children_textfield", edit=True, text="");'
              'parent = cmds.textFieldButtonGrp("reparent_parent_textfield", edit=True, text="<World>");'
              'sparseKeys = cmds.checkBoxGrp("reparent_sparseKeys_chkbox", edit=True, value1=0);' )

children_btn_cmd = ( 'import maya.cmds as cmds;\n'
                     'sel = cmds.ls(sl=True, long=True);\n'
                     'if len(sel) > 0:\n'
                     '\tchildren = ", ".join(sel);\n'
                     '\tcmds.textFieldButtonGrp("reparent_children_textfield", edit=True, text=children);\n')

parent_btn_cmd = ( 'import maya.cmds as cmds;\n'
                   'sel = cmds.ls(sl=True, long=True);\n'
                   'if len(sel) > 0:\n'
                   '\tcmds.textFieldButtonGrp("reparent_parent_textfield", edit=True, text=sel[0]);\n')


# used to clean up the UI string before sending to reparent.main(...).
def splitChildrenString(children):
    result = list()
    for child in children.split(','):
        child = child.strip()
        if cmds.objExists(child):
            result.append(child)
    if len(result) == 0:
        msg = 'No valid children were found.'
        raise ValueError, msg
    return result


# used to clean up the UI string before sending to reparent.main(...).
def splitParentString(parent):
    result = str()
    if parent.lower() == '<world>':
        result = None
    else:
        if cmds.objExists(parent):
            result = parent
        else:
            msg = 'Parent object is not valid, it does not exist.'
            raise ValueError, msg
    return result


def gui():

    # window
    window_name = 'reparent_window'
    if cmds.window( window_name, exists=1 ) == 1:
        cmds.deleteUI( window_name, window=1 )

    # make GUI
    window_width = 400
    window = cmds.window( window_name,
                          title='Reparent Window',
                          menuBar=0,
                          resizeToFitChildren=1,
                          width=window_width, height=1 )

    # main layouts
    main_layout = cmds.rowColumnLayout(
        columnWidth=[ (1, window_width) ],
        numberOfColumns=1 )

    cmds.text(label='')

    # Children
    cmds.textFieldButtonGrp('reparent_children_textfield',
                            label="Children", text="",
                            columnWidth3=(int(window_width*(0.2)),
                                          int(window_width*(0.55)),
                                          int(window_width*(0.25))),
                            buttonCommand=children_btn_cmd,
                            buttonLabel="Get Selection")

    # Parent
    cmds.textFieldButtonGrp('reparent_parent_textfield',
                            label="Parent", text="<World>",
                            columnWidth3=(int(window_width*(0.2)),
                                          int(window_width*(0.55)),
                                          int(window_width*(0.25))),
                            buttonCommand=parent_btn_cmd,
                            buttonLabel="Get Selection")

    #
    # cmds.checkBox('reparent_sparseKeys_chkbox', label="Sparse Keys")
    cmds.checkBoxGrp('reparent_sparseKeys_chkbox',
                     label="Sparse Keys", numberOfCheckBoxes=1,
                     value1=0,
                     columnWidth2=(int(window_width*(0.2)),
                                   int(window_width*(0.8))))
    
    cmds.text(label='')
    cmds.separator()

    # reparent button
    reparent_btn = cmds.button(label="Reparent", command=reparent_cmd)

    # reset button
    reset_btn = cmds.button(label="Reset", command=reset_cmd)

    # exit button
    exit_cmd = ('import maya.cmds as cmds;'
                'cmds.deleteUI("%s", window=1)') % window_name
    exit_btn = cmds.button(label="Exit", command=exit_cmd)

    # show the window
    cmds.showWindow(window_name)

    # change the window height.
    cmds.window(window_name, e=1, height=1)

    return True


def getAnimCurves(node):
    animNodes = cmds.listConnections(node, type="animCurve")
    return animNodes


def getKeyframeTimes(animCurve, timeRange):
    if not cmds.objExists(animCurve):
        return list()
    keyTimes = cmds.keyframe(animCurve, time=timeRange, query=True, timeChange=True)
    return keyTimes


def getAttributeFromAnimCurve(node, animCurve):
    conns = cmds.listConnections(node, connections=True, source=True, destination=True)
    for i, conn in enumerate(conns):
        if conn == animCurve:
            nodeAttr = conns[i-1]
            split = nodeAttr.split('.')
            if len(split) == 2:
                return split[1]
    return None


def main(children, parent, sparseKeys=False):
    saveSel = cmds.ls(sl=True, long=True)
    
    startTime = cmds.playbackOptions(query=True, min=True)
    endTime = cmds.playbackOptions(query=True, max=True)
    timeRange = (startTime, endTime)

    for child in children:

        # Get current key times for all attributes.
        attrKeyTimes = dict()
        if sparseKeys:
            animCurves = getAnimCurves(child)
            if animCurves != None:
                for animCurve in animCurves:
                    attr = getAttributeFromAnimCurve(child, animCurve)
                    times = getKeyframeTimes(animCurve, timeRange)
                    attrKeyTimes[attr] = times

        # create locator
        locTfm = cmds.createNode('transform', name='dummy')
        locShape = cmds.createNode('locator', name='dummyShape', parent=locTfm)

        # constrain the locator to the child,
        # in translate, rotate and scale.
        cmds.pointConstraint(child, locTfm)
        cmds.orientConstraint(child, locTfm)
        cmds.scaleConstraint(child, locTfm)

        # bake the locator results
        cmds.bakeResults(locTfm,
                         time=timeRange, sampleBy=1,
                         disableImplicitControl=True,
                         preserveOutsideKeys=True,
                         sparseAnimCurveBake=True,
                         controlPoints=False, shape=False)

        # delete locator history
        cmds.delete(locTfm, constraints=True)

        # remove keyframes from child transform
        cmds.cutKey(child, time=timeRange, attribute='translateX')
        cmds.cutKey(child, time=timeRange, attribute='translateY')
        cmds.cutKey(child, time=timeRange, attribute='translateZ')
        cmds.cutKey(child, time=timeRange, attribute='rotateX')
        cmds.cutKey(child, time=timeRange, attribute='rotateY')
        cmds.cutKey(child, time=timeRange, attribute='rotateZ')
        cmds.cutKey(child, time=timeRange, attribute='scaleX')
        cmds.cutKey(child, time=timeRange, attribute='scaleY')
        cmds.cutKey(child, time=timeRange, attribute='scaleZ')

        # if the parent is "" (the world), parent to the world...
        if ((parent == None) or (parent == '')):
            child = cmds.parent(child, world=True)
        else:
            # else, constrain the child to the parent,
            # delete the history, and parent the child under
            # the parent.
            if cmds.objExists(parent):
                cmds.pointConstraint(parent, child)
                cmds.orientConstraint(parent, child)
                cmds.scaleConstraint(parent, child)

                cmds.delete(child, constraints=True)

                child = cmds.parent(child, parent)

        # constrain the child to the locator,
        cmds.pointConstraint(locTfm, child)
        cmds.orientConstraint(locTfm, child)
        cmds.scaleConstraint(locTfm, child)

        # bake the child's results
        cmds.bakeResults(child,
                         time=(startTime, endTime), sampleBy=1,
                         disableImplicitControl=True,
                         preserveOutsideKeys=True,
                         sparseAnimCurveBake=True,
                         controlPoints=False, shape=False)

        # delete the child's history
        cmds.delete(child, constraints=True)

        # delete the locator.
        cmds.delete(locTfm)

        # remove keys that were not there before, "sparse keys".
        if sparseKeys:
            for attr in attrKeyTimes:
                keyTimes = attrKeyTimes[attr]
                removeTimes = list()
                for time in range(int(startTime), int(endTime)+1):
                    if time not in keyTimes:
                        cmds.cutKey(child, time=(time,time), attribute=attr,
                                    clear=True, option='keys')

    # restore selection
    tmp = saveSel
    saveSel = list()
    for s in tmp:
        if cmds.objExists(s):
            saveSel.append(s)
    if len(saveSel) > 0:
        cmds.select(saveSel, replace=True)
    return True
